/*
 * newio.c: This is some handy stuff to deal with file descriptors in a way
 * much like stdio's FILE pointers 
 *
 * IMPORTANT NOTE:  If you use the routines here-in, you shouldn't switch to
 * using normal reads() on the descriptors cause that will cause bad things
 * to happen.  If using any of these routines, use them all 
 *
 * Written By Michael Sandrof
 *
 * Copyright(c) 1990 
 *
 * See the COPYRIGHT file, or do a HELP IRCII COPYRIGHT 
 */

#if 0
static	char	rcsid[] = "@(#)$Id: newio.c,v 1.24 1995/01/15 01:13:52 mrg Exp $";
#endif

#include "irc.h"
#include "ircaux.h"
#include "output.h"
#include <sys/ioctl.h>
#ifdef HAVE_SYS_FILIO_H
#include <sys/filio.h>
#endif

#define IO_BUFFER_SIZE 512

#ifdef HAVE_SYSCONF
# define IO_ARRAYLEN sysconf(_SC_OPEN_MAX)
#else
# ifdef FD_SETSIZE
#  define IO_ARRAYLEN FD_SETSIZE
# else
#  define IO_ARRAYLEN NFDBITS
# endif
#endif

typedef	struct	myio_struct
{
	char		*buffer;
	size_t		buffer_size;
	unsigned 	read_pos,
			write_pos;
}           MyIO;

static	struct	timeval	right_away = { 0L, 0L };
static	MyIO	**io_rec;

static	struct	timeval	dgets_timer;
static	struct	timeval	timer = { 0, 0 };
	int	dgets_errno = 0;

/*
 * Get_pending_bytes: What do you think it does?
 */
size_t get_pending_bytes (int fd)
{
	if (fd >= 0 && io_rec[fd] && io_rec[fd]->buffer)
		return strlen(io_rec[fd]->buffer);

	return 0;
}

static	void	init_io (void)
{
	static	int	first = 1;

	if (first)
	{
		int	c, max_fd = IO_ARRAYLEN;

		io_rec = (MyIO **)new_malloc(sizeof(MyIO *) * max_fd);
		for (c = 0; c < max_fd; c++)
			io_rec[c] = (MyIO *) 0;
		first = 0;
	}
}

/*
 * All new dgets -- no more trap doors!
 */
int dgets (char *str, int des, int buffer)
{
	int	cnt = 0,
		c, s;
	fd_set	rd;
	MyIO	*ioe;

	init_io();
	ioe = io_rec[des];

	if (ioe == NULL)
	{
		ioe = io_rec[des] = (MyIO *)new_malloc(sizeof(MyIO));
		ioe->buffer_size = IO_BUFFER_SIZE;
		ioe->buffer = (char *)new_malloc(ioe->buffer_size + 2);
		ioe->read_pos = ioe->write_pos = 0;
	}

	if (ioe->read_pos == ioe->write_pos)
	{
		ioe->read_pos = ioe->write_pos = 0;
		ioe->buffer[0] = 0;
	}

	if (!strchr(ioe->buffer + ioe->read_pos, '\n'))
	{
		if (ioe->read_pos)
		{
			char *tmp;
			int len = strlen(ioe->buffer + ioe->read_pos);

			tmp = alloca(len + 1);
			strcpy(tmp, ioe->buffer + ioe->read_pos);
			strcpy(ioe->buffer, tmp);
			ioe->read_pos = 0;
			ioe->write_pos = len;
		}

		/*
		 * Dont try to read into a full buffer.
		 */
		if (ioe->write_pos >= ioe->buffer_size)
		{
			yell("***XXX*** Buffer for des [%d] is filled! ***XXX***", des);
			dgets_errno = ENOMEM; /* Cough */
			return -1;
		}

		FD_ZERO(&rd);
		FD_SET(des, &rd);

		s = select(des + 1, &rd, 0, 0, &timer);
		if (s < 0)
		{
			strcpy(str, empty_string);
			dgets_errno = errno;
			return -1;
		} 
		else if (s != 0)
		{
			int nbytes;
			ioctl(des, FIONREAD, &nbytes);
			c = read(des, ioe->buffer + ioe->write_pos,
				      ioe->buffer_size - ioe->write_pos);
			if (x_debug & DEBUG_INBOUND) 
			    yell("FD [%d], should [%d] did [%d]", des, nbytes, c);
			if (c <= 0)
			{
				strcpy(str, empty_string);
				dgets_errno = (c == 0) ? -1 : errno;
				return -1;
			}
			ioe->buffer[ioe->write_pos + c] = 0;
			ioe->write_pos += c;
		}
	}

	dgets_errno = 0;

	/*
	 * Check to make sure we have a newline if we're buffered.
	 */
	if (buffer && !strchr(ioe->buffer + ioe->read_pos, '\n'))
		return 0;

	/*
	 * Copy the next line out of the buffer
	 */
	while (ioe->read_pos < ioe->write_pos)
	{
		if (((str[cnt] = ioe->buffer[ioe->read_pos++])) == '\n')
			break;
		cnt++;
	}

	/*
	 * Terminate it
	 */
	str[cnt + 1] = 0;

	/*
	 * If we end in a newline, then all is well.
	 * Otherwise, we're unbuffered, tell the caller.
	 */
	if (str[cnt] == '\n')
		return cnt;
	else
		return 0;
}

/*
 * new_select: works just like select(), execpt I trimmed out the excess
 * parameters I didn't need.  
 */
int new_select (fd_set *rd, fd_set *wd, struct timeval *timeout)
{
		int	i,
			set = 0;
		fd_set 	new;
	struct timeval	thetimeout;
	struct timeval *newtimeout = &thetimeout;
		int	max_fd = -1;
	static	int	num_fd = 0;

	if (!num_fd)
	{
		num_fd = IO_ARRAYLEN; /* why do it a zillion times? */
		if (num_fd > FD_SETSIZE)
			num_fd = FD_SETSIZE;
	}

	if (timeout)
		memmove(newtimeout, timeout, sizeof(struct timeval));
	else
		newtimeout = NULL;

	init_io();
	FD_ZERO(&new);
	for (i = 0; i < num_fd; i++)
	{
		if (i > max_fd && ((rd && FD_ISSET(i, rd)) || (wd && FD_ISSET(i, wd))))
			max_fd = i;
		if (io_rec[i])
		{
			if (io_rec[i]->read_pos < io_rec[i]->write_pos)
			{
				FD_SET(i, &new);
				set = 1;
			}
		}
	}
	if (set)
	{
		set = 0;
		if (select(max_fd + 1, rd, wd, NULL, &right_away) <= 0)
			FD_ZERO(rd);
		for (i = 0; i < num_fd; i++)
		{
			if ((FD_ISSET(i, rd)) || (FD_ISSET(i, &new)))
			{
				set++;
				FD_SET(i, rd);
			}
			else
				FD_CLR(i, rd);
		}
		return (set);
	}
	return (select(max_fd + 1, rd, wd, NULL, newtimeout));
}

/* new_close: works just like close */
void new_close (int des)
{
	new_free((char **)&(io_rec[des]));
	close(des);
}

/* set's socket options */
void set_socket_options (int s)
{
	int	opt = 1;
	int	optlen = sizeof(opt);
#ifndef NO_STRUCT_LINGER
	struct linger	lin;

	lin.l_onoff = lin.l_linger = 0;
	(void) setsockopt(s, SOL_SOCKET, SO_LINGER, (char *)&lin, optlen);
#endif

	(void) setsockopt(s, SOL_SOCKET, SO_REUSEADDR, (char *)&opt, optlen);
	opt = 1;
	(void) setsockopt(s, SOL_SOCKET, SO_KEEPALIVE, (char *)&opt, optlen);

#if 0
	/* This is waiting for nonblock-aware code */
	info = fcntl(fd, F_GETFL, 0);
	info |= O_NONBLOCK;
	fcntl(fd, F_SETFL, info);
#endif
}
